/**
 * Copyright (C) 2009 Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/

package org.ow2.orchestra.definition.activity;

import java.io.Serializable;
import java.util.Map;

import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.definition.element.InboundMessageElement;
import org.ow2.orchestra.facade.data.def.ActivityDefinitionData;
import org.ow2.orchestra.facade.data.runtime.ActivityInstanceData;
import org.ow2.orchestra.facade.runtime.ActivityType;
import org.ow2.orchestra.facade.uuid.ActivityDefinitionUUID;
import org.ow2.orchestra.facade.uuid.ActivityInstanceUUID;
import org.ow2.orchestra.facade.uuid.IdFactory;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.util.EnvTool;

/**
 * @author S. Ali Tokmen
 */
public abstract class AbstractActivity implements Serializable {

  private static final long serialVersionUID = -2049270479220106510L;

  /** For persistence */
  protected long dbid;

  protected ActivityDefinitionUUID uuid;

  /** name attribute present for all activities. */
  protected String name = null;

  public long getDbid() {
    return this.dbid;
  }

  public ActivityDefinitionUUID getUUID() {
    return this.uuid;
  }

  public void setUUID(final ActivityDefinitionUUID uuid) {
    this.uuid = uuid;
  }

  public String getName() {
    return this.name;
  }

  public void setName(final String name) {
    this.name = name;
  }

  public abstract ActivityType getType();

  /**
   * Hook start of activity.
   *
   * @param bpelExecution
   */
  public void beforeRunning(final BpelExecution bpelExecution) {
    final ActivityInstanceUUID activityInstanceUUID =
      IdFactory.getNewActivityInstanceUUID(bpelExecution.getProcessInstance().getProcessInstanceUUID());
    bpelExecution.setActivityInstanceUUID(activityInstanceUUID);

    EnvTool.getRecorder().recordActivityStarted(this.createRuntimeData(bpelExecution));

  }

  /**
   * @return a definition data
   */
  public ActivityDefinitionData createDefinitionData(final BpelProcess bpelProcess) {
    return new ActivityDefinitionData(bpelProcess.getProcessDefinitionData(), this.uuid, null,
      this.getType(), this.name);
  }

  /**
   * @param bpelExecution the execution
   * @return an instance data
   */
  public final ActivityInstanceData createRuntimeData(final BpelExecution bpelExecution) {
    return new ActivityInstanceData(this.createDefinitionData(bpelExecution.getProcessDefinition()), bpelExecution
      .getProcessInstance().getProcessInstanceData(), bpelExecution.getActivityInstanceUUID(), bpelExecution
      .getEncloserActivityInstanceUUID());
  }

  /**
   * Hook end of activity.
   *
   * @param bpelExecution
   */
  public void afterRunned(final BpelExecution bpelExecution) {
    EnvTool.getRecorder().recordActivityEnded(this.createRuntimeData(bpelExecution));
  }

  /**
   * piece of Java code that is to be included in a process as node behaviour or
   * as a hidden listener to process events.
   *
   * <p>
   * Activities can be used to implement the behaviour of nodes, in which case
   * this behaviour is associated to a graphical node in the diagram, or they
   * can be added as events, in that case, they are being hidden from the
   * diagram.
   * </p>
   *
   */
  public abstract void execute(BpelExecution execution) throws Exception;

  /**
   * handles an external trigger.
   *
   * <p>
   * An external trigger that comes into an execution through one of the
   * {@link org.ow2.orchestra.pvm.Execution#signal()} methods, will be delegated to the node in which
   * the execution is positioned when it receives the external trigger.
   * </p>
   *
   * <p>
   * The signal method implements how the node will react on that signal. For
   * example, the outgoing transition could be taken that corresponds with the
   * given signal.
   * </p>
   *
   * @param execution
   *          the {@link org.ow2.orchestra.pvm.Execution} for which the signal is given
   *
   * @param signalName
   *          is an abstract text that can be associated with a signal. this
   *          corresponds to e.g. a method name in a java class interface. The
   *          implementation can decide e.g. to use the signal to identify the
   *          outgoing transition.
   *
   * @param parameters
   *          is extra information that can be provided with a signal. In this
   *          way, it is somewhat similar to the parameters that can be fed into
   *          a method invocation through reflection.
   *
   * @throws Exception
   *           to indicate any kind of failure. Note that exceptions are
   *           considered non recoverable. After an Exception, the execution
   *           should not be used any more and if this is during a transaction,
   *           the transaction should be rolled back.
   */
  public abstract void signal(
      final BpelExecution execution,
      final String signal,
      final Map<String, Object> parameters);

  @Override
  public int hashCode() {
    return this.uuid.hashCode();
  }

  @Override
  public boolean equals(final Object obj) {
    if (obj == null) {
      return false;
    }
    if (!(obj instanceof AbstractActivity)) {
      return false;
    }
    final AbstractActivity other = (AbstractActivity) obj;
    return other.getUUID().equals(this.uuid);
  }

  public boolean isInboundMessageElement() {
    return this instanceof InboundMessageElement;
  }
}
