/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.definition.activity;

import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.ow2.orchestra.definition.Node;
import org.ow2.orchestra.definition.element.Else;
import org.ow2.orchestra.definition.element.Elseif;
import org.ow2.orchestra.facade.runtime.ActivityType;
import org.ow2.orchestra.lang.Expression;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.runtime.ExecutionState;
import org.ow2.orchestra.util.Misc;

/**
 * the runtime behaviour of the if activity.
 *
 * @author Tom Baeyens, Charles Souillard
 */
public class If extends AbstractBpelActivity {

  /** serialVersionUID. */
  private static final long serialVersionUID = 1L;
  private static final Logger LOG = Logger.getLogger(If.class.getName());

  protected Expression condition;
  protected Node activity;
  protected List<Elseif> elseifs;
  protected Else elseField;

  public If() {
    super();
  }
  /**
   * business logic.
   *
   * @param execution execution
   */
  @Override
  public void executeActivity(final BpelExecution execution) {
    final boolean ifResult = this.condition.getEvaluator().evaluateBoolean(execution);
    if (ifResult) {
      Misc.fastDynamicLog(If.LOG, Level.FINE, "if condition : %s evaluated to true", this.condition);

      execution.createExecution().execute(this.activity);
    } else {
      // manage elseifs
      for (final Elseif elseif : this.elseifs) {
        final boolean elseifResult = elseif.getCondition().getEvaluator().evaluateBoolean(execution);
        if (elseifResult) {
          Misc.fastDynamicLog(If.LOG, Level.FINE, "elseif condition : %s evaluated to true", elseif.getCondition());

          execution.createExecution().execute(elseif.getActivity());
          return;
        }
      }
      // execute else
      if (this.elseField != null) {
        Misc.fastDynamicLog(If.LOG, Level.FINE, "Executing else...");

        execution.createExecution().execute(this.elseField.getActivity());
      } else {
        // no branch of the if is executed
        // finish activity execution
        this.afterRunned(execution);
      }
    }
  }

  /**
   * Signal method of the NodeBehaviour.
   *
   * @param execution execution
   * @param signal signal
   */
  @Override
  public void signal(final BpelExecution execution, final String signal, final Map<String, Object> parameters) {
    if (signal == null) {
      // this is the child execution returning
      // end child execution
      execution.end(ExecutionState.ended);

      final BpelExecution ifExecution = execution.getParent();
      // remove finished execution from parent
      ifExecution.removeExecution(execution);

      // signal must be non-null to make the distinction with child
      // executions
      // in the first line of this method
      ifExecution.signal("finished");
    } else if (signal.equals("finished")) {
      this.afterRunned(execution);
      // execution.proceed();
    }
  }
  public Expression getCondition() {
    return this.condition;
  }
  public void setCondition(final Expression condition) {
    this.condition = condition;
  }
  public List<Elseif> getElseifs() {
    return this.elseifs;
  }
  public void setElseifs(final List<Elseif> elseifs) {
    this.elseifs = elseifs;
  }
  public Else getElse() {
    return this.elseField;
  }
  public void setElse(final Else else1) {
    this.elseField = else1;
  }
  public Node getActivity() {
    return this.activity;
  }
  public void setActivity(final Node activity) {
    this.activity = activity;
  }

  @Override
  public ActivityType getType() {
    return ActivityType.IF;
  }

}
