/**
 * Bull SAS / OW2 Orchestra
 * Copyright (C) 2007 Bull S.A.S, and individual contributors as indicated
 * by the @authors tag.
 *
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.definition.element;

import java.io.Serializable;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.wsdl.Port;
import javax.wsdl.Service;
import javax.wsdl.extensions.ExtensibilityElement;
import javax.wsdl.extensions.soap.SOAPAddress;
import javax.xml.namespace.QName;

import org.ow2.orchestra.util.BpelUtil;
import org.ow2.orchestra.util.Misc;
import org.ow2.orchestra.wsdl.WsdlsInfos;
import org.w3c.dom.Document;

/**
 * @author Goulven Le Jeune, Charles Souillard, Candy Chlad, Stanislas Giraudet
 *         De Boudemange
 */
public class PartnerLink implements Serializable {
  private static final Logger LOG = Logger.getLogger(PartnerLink.class.getName());

  /** serialVersionUID. */
  private static final long serialVersionUID = 1L;

  protected long dbid;

  /** name attribute of this partnerLink. */
  protected String name;

  /**
   * portType QName of the PartnerLinkType associated to myRole of this
   * partnerLink.
   */
  protected QName myRolePortTypeQName;

  /**
   * portType QName of the PartnerLinkType associated to partnerRole of this
   * partnerLink.
   */
  protected QName partnerRolePortTypeQName;

  /** intitializePartnerRole Attribute. */
  protected boolean initializePartnerRole = false;

  protected Document defaultPartnerRoleEndPointReference;

  protected Document defaultMyRoleEndPointReference;


  public String getName() {
    return this.name;
  }

  public void setName(final String name) {
    this.name = name;
  }

  public void setInitializePartnerRole(final boolean initializePartnerRole) {
    this.initializePartnerRole = initializePartnerRole;
  }

  public QName getMyRolePortTypeQName() {
    return this.myRolePortTypeQName;
  }

  public void setMyRolePortTypeQName(final QName myRolePortTypeQName) {
    this.myRolePortTypeQName = myRolePortTypeQName;
  }

  public QName getPartnerRolePortTypeQName() {
    return this.partnerRolePortTypeQName;
  }

  public void setPartnerRolePortTypeQName(final QName partnerRolePortTypeQName) {
    this.partnerRolePortTypeQName = partnerRolePortTypeQName;
  }

  public boolean getInitializePartnerRole() {
    return this.initializePartnerRole;
  }


  /**
   * @return the defaultMyRoleEndPointReference
   */
  public Document getDefaultMyRoleEndPointReference() {
    return this.defaultMyRoleEndPointReference;
  }


  /**
   * @return the defaultPartnerRoleEndPointReference
   */
  public Document getDefaultPartnerRoleEndPointReference() {
    return this.defaultPartnerRoleEndPointReference;
  }

  /**
   * @param wsdlInfos
   */
  public void getDefaultEndpoints(final WsdlsInfos wsdlInfos) {
    this.initializePartnerRole(wsdlInfos);
    this.initializeMyRole(wsdlInfos);
  }

  /**
   * Initialize partnerRole endpoint reference to the defaultPort value.
   *
   * @param execution
   *            execution to which this partnerLinkRuntime belongs to
   */
  @SuppressWarnings("unchecked")
  private void initializePartnerRole(final WsdlsInfos wsdlsInfos) {
    Port defaultPort = null;

    if (this.partnerRolePortTypeQName == null) {
      return;
    }

    final List<Port> ports = wsdlsInfos.getPorts(this.partnerRolePortTypeQName);

    final Map<Port, SOAPAddress> supported = new HashMap<Port, SOAPAddress>();
    for (final Port port : ports) {
      final List<ExtensibilityElement> elements = port.getExtensibilityElements();
      if (elements.size() > 0) {
        for (final ExtensibilityElement element : elements) {
          if (element instanceof SOAPAddress) {
            if (port.getBinding().getPortType().getQName().equals(this.partnerRolePortTypeQName)) {
              supported.put(port, (SOAPAddress) element);
            }
          }
        }
      } else {
        supported.put(port, null);
      }
    }

    SOAPAddress soapAddress = null;
    Service service = null;
    if (supported.entrySet() != null && supported.entrySet().size() > 0) {
      final Map.Entry<Port, SOAPAddress> entry = supported.entrySet().iterator().next();
      defaultPort = entry.getKey();
      soapAddress = entry.getValue();

      Misc.fastDynamicLog(PartnerLink.LOG, Level.FINEST, "Using the supported port: %s", defaultPort);

      service = wsdlsInfos.getServiceOfPort(defaultPort);
    }

    if (service != null) {
      this.defaultPartnerRoleEndPointReference = BpelUtil.getServiceRefDocument(soapAddress, service.getQName());
    }
  }

  /**
   * Initialize myRole endpoint reference to the defaultPort value.
   *
   * @param execution
   *            execution to which this partnerLinkRuntime belongs to
   */
  @SuppressWarnings("unchecked")
  private void initializeMyRole(final WsdlsInfos wsdlsInfos) {
    if (this.myRolePortTypeQName == null) {
      return;
    }

    final List<Port> ports = wsdlsInfos.getPorts(this.myRolePortTypeQName);

    if (ports.isEmpty()) {
      return;
    }
    final Port defaultPort = ports.get(0);
    final Service service = wsdlsInfos.getServiceOfPort(defaultPort);
    final List<ExtensibilityElement> elements = defaultPort.getExtensibilityElements();
    final SOAPAddress soapAddress = (SOAPAddress) BpelUtil.findExtensibilityElement(elements, SOAPAddress.class);
    this.defaultMyRoleEndPointReference = BpelUtil.getServiceRefDocument(soapAddress, service.getQName());
  }

}
