/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.jmx;

import java.util.List;

import org.ow2.orchestra.env.EnvironmentFactory;
import org.ow2.orchestra.facade.InstanceManagementAPI;
import org.ow2.orchestra.facade.data.runtime.JobData;
import org.ow2.orchestra.facade.exception.InstanceNotFoundException;
import org.ow2.orchestra.facade.exception.JobNotFoundException;
import org.ow2.orchestra.facade.uuid.ProcessInstanceUUID;
import org.ow2.orchestra.jmx.commands.ExitInstanceCommand;
import org.ow2.orchestra.jmx.commands.FindJobsWithExceptionCommand;
import org.ow2.orchestra.jmx.commands.RemoveInstanceCommand;
import org.ow2.orchestra.jmx.commands.ResumeInstanceCommand;
import org.ow2.orchestra.jmx.commands.RetryJobCommand;
import org.ow2.orchestra.jmx.commands.SuspendInstanceCommand;
import org.ow2.orchestra.services.commands.CommandService;

public class InstanceManagementAPIImpl implements InstanceManagementAPI {

  private final EnvironmentFactory environmentFactory;

  public InstanceManagementAPIImpl(final EnvironmentFactory environmentFactory) {
    this.environmentFactory = environmentFactory;
  }

  protected CommandService getCommandService() {
    return this.environmentFactory.get(CommandService.class);
  }

  /**
   * Exit the given instance.
   *
   * @param processInstanceUUID
   *          the Id of the instance to exit.
   * @throws InstanceNotFoundException
   *           thrown if the given instance to exit is not found.
   */
  public void exit(final ProcessInstanceUUID processInstanceUUID) throws InstanceNotFoundException {
    final InstanceNotFoundException infe = this.getCommandService().execute(new ExitInstanceCommand(processInstanceUUID));
    if (infe != null) {
      throw infe;
    }
  }

  /**
   * Resume a previously suspend instance of a process.
   *
   * @param processInstanceUUID
   *          the process instance UUID that identify the instance to resume.
   * @throws InstanceNotFoundException
   *           thrown if the given instance to resume is not found.
   */
  public void resume(final ProcessInstanceUUID processInstanceUUID) throws InstanceNotFoundException {
    final InstanceNotFoundException infe = this.getCommandService().execute(new ResumeInstanceCommand(processInstanceUUID));

    if (infe != null) {
      throw infe;
    }


  }

  /**
   * Suspend the given instance of a process.
   *
   * @param processInstanceUUID
   *          the process instance UUID that identify the instance to suspend.
   * @throws InstanceNotFoundException
   *           thrown if the given instance to suspend is not found.
   */
  public void suspend(final ProcessInstanceUUID processInstanceUUID) throws InstanceNotFoundException {

    final InstanceNotFoundException infe = this.getCommandService().execute(new SuspendInstanceCommand(processInstanceUUID));

    if (infe != null) {
      throw infe;
    }
  }

  /* (non-Javadoc)
   * @see org.ow2.orchestra.facade.InstanceManagementAPI#remove(org.ow2.orchestra.facade.uuid.ProcessInstanceUUID)
   */
  public boolean remove(final ProcessInstanceUUID processInstanceUUID) {
    return this.getCommandService().execute(new RemoveInstanceCommand(processInstanceUUID)).booleanValue();
  }


  /**
   * Get the list of asynchronous executions for which all the tries have failed and which will not be executed any more.
   */
  public List<JobData> findJobsWithException() {
    return this.getCommandService().execute(new FindJobsWithExceptionCommand());
  }

  /**
   * Reschedules a job (asynchronous execution).
   * The job executor will try to execute the job at most retries times.
   * If the job fails for each retry, then the job executor will not try to execute it
   * unless this method is called again for this job.
   * @param jobData the job to retry.
   * @param retries the number of retries to set for the job.
   * @throws JobNotFoundException if job does not exist.
   */
  @SuppressWarnings("boxing")
  public void retryJob(final JobData jobData, final int retries) throws JobNotFoundException {
    if (!this.getCommandService().execute(new RetryJobCommand(retries, jobData))) {
      // if command returned false, job was not found
      // throw exception
      throw new JobNotFoundException(jobData);
    }
  }
}
