package org.ow2.orchestra.persistence.db.hibernate;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.xml.namespace.QName;

import org.hibernate.Cache;
import org.hibernate.FlushMode;
import org.hibernate.HibernateException;
import org.hibernate.LockMode;
import org.hibernate.LockOptions;
import org.hibernate.Query;
import org.hibernate.ScrollableResults;
import org.hibernate.Session;
import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.definition.BpelProcess.InstanceCreationLock;
import org.ow2.orchestra.definition.element.Correlation;
import org.ow2.orchestra.definition.element.Correlation.Initiate;
import org.ow2.orchestra.definition.element.InboundMessageElement;
import org.ow2.orchestra.deployment.ActiveProcessHolder;
import org.ow2.orchestra.deployment.ProcessDeploymentCounter;
import org.ow2.orchestra.facade.uuid.ProcessDefinitionUUID;
import org.ow2.orchestra.facade.uuid.ProcessInstanceUUID;
import org.ow2.orchestra.persistence.db.RuntimeDbSession;
import org.ow2.orchestra.runtime.BpelInstance;
import org.ow2.orchestra.runtime.WaitingExecution;
import org.ow2.orchestra.runtime.WaitingExecution.WaitingExecutionIterator;
import org.ow2.orchestra.services.OperationKey;
import org.ow2.orchestra.services.PendingMessage;
import org.ow2.orchestra.services.ReceivingElement;
import org.ow2.orchestra.services.job.Job;
import org.ow2.orchestra.util.Misc;
import org.ow2.orchestra.util.ReceivingElementUtil;
import org.w3c.dom.Document;

public class RuntimeDbSessionImpl extends
    HibernateDbSession implements RuntimeDbSession {

  private static final Logger LOG = Logger.getLogger(RuntimeDbSessionImpl.class.getName());
  private static final String PROCESS_QUERY_CACHE = "processQueryCache";

  public RuntimeDbSessionImpl(final Session session) {
    this.setSession(session);
    session.setFlushMode(FlushMode.COMMIT);
  }

  public ActiveProcessHolder findActiveBpelProcess(final QName processQName) {
    final Query query = this.getSession().getNamedQuery("getActiveProcessHolder");
    query.setString("qname", processQName.toString());
    query.setCacheable(true);
    query.setCacheRegion(RuntimeDbSessionImpl.PROCESS_QUERY_CACHE);
    return (ActiveProcessHolder) query.uniqueResult();
  }

  public List<ProcessDefinitionUUID> findBpelProcesses(final QName processQName, final boolean lock) {
    if (lock) {
      final Query query = this.getSession().getNamedQuery("findBpelProcessesUUIDByQNameForLock");
      query.setString("qname", processQName.toString());
      query.setLockOptions(new LockOptions(LockMode.PESSIMISTIC_WRITE));
      // don't use cache for this request or the lock
      query.setCacheable(false);
      final List<ProcessDefinitionUUID> res = query.list();
      return res;
    }
    return this.findBpelProcesses(processQName);
  }

  private List<ProcessDefinitionUUID> findBpelProcesses(final QName processQName) {
    final Query query = this.getSession().getNamedQuery("findBpelProcessesUUIDByQName");
    query.setString("namespace", processQName.getNamespaceURI());
    query.setString("name", processQName.getLocalPart());
    query.setCacheable(true);
    return query.list();
  }

  public BpelProcess findBpelProcess(final ProcessDefinitionUUID pduuid, final boolean lock) {
    if (lock) {
      final Query query = this.getSession().getNamedQuery("findBpelProcessForLock");
      query.setString("uuid", pduuid.toString());
      query.setLockOptions(new LockOptions(LockMode.PESSIMISTIC_WRITE));
      // don't use cache for this request or the lock
      query.setCacheable(false);
      final InstanceCreationLock res = (InstanceCreationLock) query.uniqueResult();
      if (res == null) {
        return null;
      }
    }
    return this.getBpelProcess(pduuid);
  }

  public BpelProcess getBpelProcess(final ProcessDefinitionUUID pduuid) {
    final Query query = this.getSession().getNamedQuery("getBpelProcessByUUID");
    query.setString("uuid", pduuid.toString());
    query.setCacheable(true);
    final BpelProcess processDefinition = (BpelProcess) query.uniqueResult();
    return processDefinition;
  }

  public List<ProcessDefinitionUUID> getBpelProcessesUUID() {
    final Query query = this.getSession().getNamedQuery("getBpelProcesses");
    query.setCacheable(true);
    return query.list();
  }

  private static class WaitingExecutionIteratorImpl implements WaitingExecutionIterator {

    private final ScrollableResults results;
    private final Session session;
    private WaitingExecution current = null;

    public WaitingExecutionIteratorImpl(final ScrollableResults results, final Session session) {
      this.results = results;
      this.session = session;
    }
    public WaitingExecution next() {
      if (this.results == null) {
        return null;
      }
      if (this.current != null) {
        this.session.evict(this.current);
      }
      if (this.results.next()) {
        try {
          this.current = (WaitingExecution) this.results.get(0);
        } catch (final HibernateException e) {
          this.current = null;
        }
        if (this.current == null) {
          Misc.fastDynamicLog(RuntimeDbSessionImpl.LOG, Level.INFO, "WaitingExecution not found, skipping.");
          return this.next();
        }
        return this.current;
      } else {
        this.results.close();
      }
      return null;
    }
  }


  private List<Set<QName>> getPermutations(final Set<QName> firstQNames, final Set<Correlation> optionalCorrelations) {
    if (optionalCorrelations.isEmpty()) {
      return Arrays.asList(firstQNames);
    }
    final Correlation firstAvailable = optionalCorrelations.iterator().next();
    optionalCorrelations.remove(firstAvailable);

    final List<Set<QName>> result = new ArrayList<Set<QName>>();
    final Set<QName> res1  = new HashSet<QName>(firstQNames);
    res1.addAll(firstAvailable.getCorrelationSet().getProperties().keySet());
    result.addAll(this.getPermutations(res1, optionalCorrelations));
    final Set<QName> res2  = new HashSet<QName>(firstQNames);
    result.addAll(this.getPermutations(res2, optionalCorrelations));
    return result;
  }

  public WaitingExecutionIterator getWaitingExecutions(
      final ReceivingElement receivingElement,
      final Map<QName, Document> csProperties) {

    final Set<QName> requiredProperties = new HashSet<QName>();
    final Set<Correlation> optionalCorrelations = new HashSet<Correlation>();
    for (final Correlation correlation : receivingElement.getCorrelations()) {
      if (correlation.getInitiate().equals(Initiate.no)) {
        requiredProperties.addAll(correlation.getCorrelationSet().getProperties().keySet());
      } else if (correlation.getInitiate().equals(Initiate.join)) {
        optionalCorrelations.add(correlation);
      }
    }


    final List<Set<QName>> propertiesValues = this.getPermutations(requiredProperties, optionalCorrelations);

    Misc.fastDynamicLog(RuntimeDbSessionImpl.LOG, Level.INFO,
        "getWaitingExecutions:  %s possible correlations found.", propertiesValues.size());

    for (final Set<QName> propQNames : propertiesValues) {
      Misc.fastDynamicLog(RuntimeDbSessionImpl.LOG, Level.INFO,
          "getWaitingExecutions:  trying to match %s correlation properties.", propQNames.size());
      final Map<QName, Document> csProps = new HashMap<QName, Document>(csProperties);
      csProps.keySet().retainAll(propQNames);

      final int correlationKey = ReceivingElementUtil.getCorrelationPropertiesKey(csProps);

      final Query query = this.getSession().getNamedQuery("getWaitingExecutionsByCorrelationKey");
      query.setParameter("activity", receivingElement.getInboundMessageElement().getUUID().toString());
      query.setInteger("correlationKey", correlationKey);

      final ScrollableResults waitingExecutions = query.scroll();
      if (waitingExecutions.first()) {
        waitingExecutions.beforeFirst();
        return new WaitingExecutionIteratorImpl(waitingExecutions, this.getSession());
      } else {
        waitingExecutions.close();
      }
    }
    return new WaitingExecutionIteratorImpl(null, this.getSession());
  }

  public WaitingExecution getFirstWaitingExecution(final InboundMessageElement inboundMessageElement) {
    final Query query = this.getSession().getNamedQuery("getWaitingExecutions");
    query.setParameter("activity", inboundMessageElement.getUUID().toString());
    query.setMaxResults(1);
    return (WaitingExecution) query.uniqueResult();
  }

  public Collection<PendingMessage> getPendingMessages(final OperationKey operationKey) {
    final Query query = this.getSession().getNamedQuery("getPendingMessages");
    query.setString("operationKey", operationKey.toString());
    return query.list();
  }

  public List<BpelInstance> getBpelInstances() {
    final Query query = this.getSession().getNamedQuery("getBpelInstances");
    return query.list();
  }

  public BpelInstance getBpelInstance(final ProcessInstanceUUID processInstanceUUID) {
    final Query query = this.getSession().getNamedQuery("getBpelInstance");
    query.setString("processInstanceUUID", processInstanceUUID.toString());
    final BpelInstance bpelExecution = (BpelInstance) query.uniqueResult();
    return bpelExecution;
  }

  public List<BpelInstance> getBpelInstances(final QName processQName) {
    final Query query = this.getSession().getNamedQuery("getBpelInstancesOfProcess");
    query.setString("namespace", processQName.getNamespaceURI());
    query.setString("name", processQName.getLocalPart());
    return query.list();
  }

  public List<BpelInstance> getBpelInstances(final ProcessDefinitionUUID pduuid) {
    final Query query = this.getSession().getNamedQuery("getBpelInstancesOfProcessByUUID");
    query.setString("uuid", pduuid.toString());
    return query.list();
  }

  public ProcessDeploymentCounter getProcessDeploymentCounter() {
    return (ProcessDeploymentCounter) this.getSession().createCriteria(ProcessDeploymentCounter.class).uniqueResult();
  }

  public List<Job> findJobsWithException() {
    // query definition can be found at the bottom resource
    // org/ow2/orchestra/pvm/hibernate.job.hbm.xml
    final Query query = this.session.getNamedQuery("findJobsWithException");
    return query.list();
  }

  public void clearActiveProcessCache(final QName processQName) {
    final Cache cache = this.getSession().getSessionFactory().getCache();
    if (cache != null) {
      final ActiveProcessHolder activeProcessHolder = this.findActiveBpelProcess(processQName);
      if (activeProcessHolder != null) {
        cache.evictEntity(ActiveProcessHolder.class, activeProcessHolder.getDbid());
      }
      cache.evictQueryRegion(RuntimeDbSessionImpl.PROCESS_QUERY_CACHE);

    }
  }
}
