/*
 * JBoss, Home of Professional Open Source
 * Copyright 2005, JBoss Inc., and individual contributors as indicated
 * by the @authors tag. See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.ow2.orchestra.persistence.tx;

import java.util.logging.Level;
import java.util.logging.Logger;

import org.ow2.orchestra.env.Environment;
import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.services.commands.Command;
import org.ow2.orchestra.services.commands.Interceptor;
import org.ow2.orchestra.util.Misc;

/**
 * calls setRollbackOnly on the transaction in the environment in case an
 * exception occurs during execution of the command.
 *
 * @author Tom Baeyens
 */
public class StandardTransactionInterceptor extends Interceptor {

  private static final Logger LOG = Logger.getLogger(StandardTransactionInterceptor.class.getName());

  public <T> T execute(final Command<T> command) {
    final Environment environment = Environment.getCurrent();
    if (environment == null) {
      throw new OrchestraRuntimeException("no environment for managing hibernate transaction");
    }

    final StandardTransaction standardTransaction = environment.get(StandardTransaction.class);
    if (standardTransaction == null) {
      throw new OrchestraRuntimeException("no standard-transaction in environment");
    }

    standardTransaction.begin();

    try {
      return this.next.execute(command);

    } catch (final RuntimeException e) {
      // Before throwing, trace the exception since it'll be lost of the finally
      // clause also throws an exception
      Misc.fastDynamicLog(StandardTransactionInterceptor.LOG, Level.SEVERE,
          "Failed executing command %s: %s %s", command, e, Misc.getStackTraceFrom(e));

      standardTransaction.setRollbackOnly();
      throw e;
    } finally {
      standardTransaction.complete();
    }
  }
}
