/**
 * Copyright (C) 2010  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.runtime;

import java.util.Date;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.definition.activity.AbstractActivity;
import org.ow2.orchestra.env.Environment;
import org.ow2.orchestra.facade.data.runtime.ProcessInstanceData;
import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.facade.runtime.ActivityState;
import org.ow2.orchestra.facade.runtime.ActivityType;
import org.ow2.orchestra.facade.uuid.ProcessDefinitionUUID;
import org.ow2.orchestra.facade.uuid.ProcessInstanceUUID;
import org.ow2.orchestra.persistence.db.MessageSession;
import org.ow2.orchestra.services.itf.ProcessRepository;
import org.ow2.orchestra.services.itf.Replier;
import org.ow2.orchestra.services.job.CheckHasMessageJob;
import org.ow2.orchestra.services.job.Job;
import org.ow2.orchestra.services.job.JobImpl;
import org.ow2.orchestra.services.job.Timer;
import org.ow2.orchestra.util.EnvTool;
import org.ow2.orchestra.util.Misc;


/**
 * @author Guillaume Porcher
 *
 */
public class BpelInstance extends BpelExecution {

  /** log. */
  private static final Logger LOG = Logger.getLogger(BpelExecution.class.getName());

  protected ProcessInstanceUUID processInstanceUUID;

  protected Date parentInstanceStartTime;

  /**
   * State of the process instance.
   */
  protected ActivityState instanceState = null;

  protected Set<Replier> repliers = new HashSet<Replier>();

  /** process definition is not persisted, only its uuid */
  protected ProcessDefinitionUUID processUUID;
  protected transient BpelProcess bpelProcess;

  // Constructor is empty because hibernate will call this constructor
  // each time the object is fetched from the database
  public BpelInstance() { }

  public Set<Replier> getRepliers() {
    return this.repliers;
  }

  public void addReplier(final Replier replier) {
    this.getRepliers().add(replier);
  }

  public ProcessInstanceUUID getProcessInstanceUUID() {
    return this.processInstanceUUID;
  }

  public void setProcessInstanceUUID(final ProcessInstanceUUID processInstanceUUID) {
    this.processInstanceUUID = processInstanceUUID;
  }

  public ProcessInstanceData getProcessInstanceData() {
    return new ProcessInstanceData(this.getProcessDefinition().getProcessDefinitionData(), this.getProcessInstanceUUID(),
        this.getProcessInstanceState());
  }

  /**
   * Gets the state of the instance.
   *
   * @return whether the instance is suspended or not.
   */
  public boolean isInstanceSuspended() {
    return ActivityState.SUSPENDED.equals(this.getProcessInstanceState());
  }

  /**
   * Suspends this instance.
   */
  public void suspendInstance() {
    if (ActivityState.RUNNING.equals(this.getProcessInstanceState())) {
      BpelInstance.suspend(this);
      this.setProcessInstanceState(ActivityState.SUSPENDED);
      EnvTool.getRecorder().recordInstanceUpdated(this.getProcessInstanceData());
    }
  }

  /**
   * Resumes this instance.
   */
  public void resumeInstance() {
    if (ActivityState.SUSPENDED.equals(this.getProcessInstanceState())) {
      BpelInstance.resume(this);
      this.setProcessInstanceState(ActivityState.RUNNING);
      EnvTool.getRecorder().recordInstanceUpdated(this.getProcessInstanceData());
    }
  }

  public ActivityState getProcessInstanceState() {
    return this.instanceState;
  }

  public void setProcessInstanceState(final ActivityState state) {
    this.instanceState = state;
  }

  @Override
  protected void removeExecution() {
    super.removeExecution();
    // This is the root BpelExecution object of the process instance
    this.setProcessInstanceState(ActivityState.EXITED);
  }

  /**
   * The process of this execution is guaranteed to be a {@link BpelProcess}.
   */
  @Override
  public BpelProcess getProcessDefinition() {
    if (this.bpelProcess == null) {
      this.bpelProcess = Environment.getFromCurrent(ProcessRepository.class).getProcess(this.processUUID);
    }
    return this.bpelProcess;
  }

  // construction /////////////////////////////////////////////////////////////

  public void initializeProcessInstance(final BpelProcess executionProcessDefinition) {
    this.processUUID = executionProcessDefinition.getUUID();
    this.setNode(executionProcessDefinition.getInitial());
    this.processInstance = this;
    this.state = ExecutionState.created;
    this.parentInstanceStartTime = new Date();
  }

  // execution method : start /////////////////////////////////////////////////

  public void begin() {
    if (!ExecutionState.created.equals(this.state)) {
      throw new OrchestraRuntimeException(this.toString() + " is already begun: " + this.state);
    }
    this.setNode(this.getProcessDefinition().getInitial());
    this.state = ExecutionState.active;
    if (this.getNode() != null) {
      this.performAtomicOperation(BpelExecution.EXECUTE_NODE);
    }
  }

  /**
   * @param startTime the startTime to set
   */
  public void setParentInstanceStartTime(final Date startTime) {
    this.parentInstanceStartTime = startTime;
  }

  /**
   * @return the startTime
   */
  public Date getParentInstanceStartTime() {
    return this.parentInstanceStartTime;
  }

  public ProcessDefinitionUUID getProcessDefinitionUUID() {
    return this.processUUID;
  }

  /**
   * Recursively resumes all terminal children of the instance.
   *
   * @param execution
   *          the node on which the children have to be suspended.
   */
  private static void resume(final BpelExecution execution) {
    // Resume children
    for (final BpelExecution n : execution.getExecutions()) {
      BpelInstance.resume(n);
    }
    if (execution.isSuspended()) {
      // Resume execution
      // Check state to set
      if (execution.getResumeState() == null) {
        throw new OrchestraRuntimeException("unable to resume " + execution + ": resume state is not defined");
      }
      // If execution becomes active and is an inbound message activity,
      // check if a new message arrived when the execution was suspended
      // If execution is on a Pick, check the alarms are not fired before checking for messages.
      if (execution.getResumeState().equals(ExecutionState.active)) {
        final AbstractActivity activity = execution.getNode().getBehaviour();
        // check for the Alarms in a Pick Activity. If an alarm is reached, other
        // onMessage activities must be removed from their associated receiver.
        if (activity.getType().equals(ActivityType.PICK)) {

          boolean removeOnMessage = false;
          final Iterator<JobImpl<?>> nodeJobs = execution.getJobs().iterator();

          // check if an alarm is already reached
          while (nodeJobs.hasNext() && !removeOnMessage) {
            final Job j = nodeJobs.next();
            if (j instanceof Timer) {
              // if the job is a timer, compare it with the current time.
              if (j.getDueDate().before(new Date(System.currentTimeMillis()))) {
                removeOnMessage = true;
              }
            } else if (j instanceof CheckHasMessageJob) {
              nodeJobs.remove();
            }
          }

          if (removeOnMessage) {
            // remove all OnMessage for this Pick activity from the waiting list.
            execution.resetWaitingExecution();
          } else {
            // check if a new message has arrived
            Environment.getFromCurrent(MessageSession.class).send(new CheckHasMessageJob(execution));
          }
        } else if (activity.isInboundMessageElement()) {
          // check if a new message has arrived
          execution.removeJobs();
          Environment.getFromCurrent(MessageSession.class).send(new CheckHasMessageJob(execution));
        }
      }
      // Resume execution
      execution.setState(execution.getResumeState());
      // Resume jobs
      for (final Job j : execution.getJobs()) {
        final JobImpl<?> job = (JobImpl<?>) j;
        job.setSuspended(false);
        if (job.getLockOwner() != null) {
          job.setLockOwner(null);
          job.setLockExpirationTime(null);
        }
      }
      Misc.fastDynamicLog(BpelInstance.LOG, Level.FINE, "Execution %s now running.", execution);
    } else {
      Misc.fastDynamicLog(BpelInstance.LOG, Level.FINE, "Execution %s is already running.", execution);
    }
  }

  /**
   * Recursively suspends all terminal children of the instance.
   *
   * @param execution
   *          the execution on which the children have to be suspended.
   */
  private static void suspend(final BpelExecution execution) {
    // Suspend children
    if (!execution.getExecutions().isEmpty()) {
      for (final BpelExecution n : execution.getExecutions()) {
        BpelInstance.suspend(n);
      }
    }
    if (!execution.isSuspended()) {
      // Suspend execution
      execution.setResumeState(execution.getState());
      execution.setState(ExecutionState.suspended);
      // Suspend jobs
      for (final Job j : execution.getJobs()) {
        final JobImpl<?> job = (JobImpl<?>) j;
        job.setSuspended(true);
        if (job.getLockOwner() != null) {
          job.setLockOwner(null);
          job.setLockExpirationTime(null);
        }
      }
      Misc.fastDynamicLog(BpelInstance.LOG, Level.FINE, "Execution %s suspended.", execution);
    } else {
      Misc.fastDynamicLog(BpelInstance.LOG, Level.FINE, "Execution %s is already suspended.", execution);
    }
  }
}
