/**
 * Copyright (C) 2008 Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/

/**
 *
 */
package org.ow2.orchestra.services;

import java.util.Date;
import java.util.List;

import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.env.Environment;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.services.commands.Command;
import org.ow2.orchestra.services.itf.Repository;
import org.ow2.orchestra.util.EnvTool;
import org.ow2.orchestra.var.MessageVariable;

public class AsyncAssociateMessage implements Command<BpelExecution> {

  private static final long serialVersionUID = 6701148536558031859L;

  protected OperationKey operationKey;
  protected MessageVariable incomingMessage;
  protected MessageCarrier messageCarrier;
  protected boolean lock;
  protected Date parentInstanceStartDate;
  /**
   * @param operationKey
   * @param incomingMessage
   * @param messageCarrier
   * @param lock
   */
  public AsyncAssociateMessage(
      final OperationKey operationKey,
      final MessageVariable incomingMessage,
      final MessageCarrier messageCarrier,
      final boolean lock,
      final Date parentInstanceStartDate) {
    this.operationKey = operationKey;
    this.incomingMessage = incomingMessage;
    this.messageCarrier = messageCarrier;
    this.lock = lock;
    this.parentInstanceStartDate = parentInstanceStartDate;
  }

  public BpelExecution execute(final Environment environment) {
    final Repository repository = EnvTool.getRepository();

    final List<BpelProcess> bpelProcesses;
    // First create pending message from incoming message.
    // This message will be stored in Receiver until it is successfully used in an instance.
    if (this.lock) {
      // lock process
      bpelProcesses = repository.lockProcesses(this.operationKey.getProcessQName());
    } else {
      bpelProcesses = repository.getProcesses(this.operationKey.getProcessQName());
    }
    final PendingMessage pendingMessage =
      Receiver.storeIncomingMessage(this.incomingMessage, this.messageCarrier, this.operationKey);

    // Try to associate pending message to a running instance
    for (final BpelProcess bpelProcess : bpelProcesses) {
      // try to associate pendingMessage
      final ExecElementToSignal newExecElt =
        Receiver.associatePendingMessage(pendingMessage, bpelProcess, this.operationKey, this.parentInstanceStartDate);
      if (newExecElt != null) {
        // an execution has been found
        // signal it in the same transaction.
        SignalExecUtil.execute(
            environment,
            newExecElt.getBpelExecution(),
            newExecElt.getReceivingElement(),
            newExecElt.getPendingMessage());
        return newExecElt.getBpelExecution();
      }
    }

    final BpelProcess activeBpelProcess = repository.getActiveProcess(this.operationKey.getProcessQName());
    // Try to create a new process instance by using the process not retired.
    if (activeBpelProcess != null) {
      // try to create new instance
      final ExecElementToSignal newExecElt =
        Receiver.startNewInstance(pendingMessage, activeBpelProcess, this.operationKey, this.parentInstanceStartDate);
      if (newExecElt != null) {
        // an execution has been found
        // signal it in the same transaction.
        SignalExecUtil.execute(
            environment,
            newExecElt.getBpelExecution(),
            newExecElt.getReceivingElement(),
            newExecElt.getPendingMessage());
        return newExecElt.getBpelExecution();
      }
    }
    // Message has not been associated.
    Receiver.storePendingMessage(pendingMessage);
    return null;
  }

}
