/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.services;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.xml.namespace.QName;

import org.ow2.orchestra.env.Environment;
import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.persistence.Transaction;
import org.ow2.orchestra.persistence.tx.StandardResource;
import org.ow2.orchestra.persistence.tx.StandardTransaction;
import org.ow2.orchestra.util.Misc;
import org.ow2.orchestra.var.MessageVariable;

public final class MessageCarrierRepository {

  private static final Logger LOG = Logger.getLogger(MessageCarrierRepository.class.getName());

  private static class TransactionnalMap<K,V> implements Map<K, V> {

    private final Map<K,V> mapToCopy;

    private Map<K,V> map = new HashMap<K, V>();
    private final List<K> addedEntries = new ArrayList<K>();
    private final List<K> removedEntries = new ArrayList<K>();

    public TransactionnalMap(final Transaction tx, final Map<K, V> mapToCopy) {
      synchronized (mapToCopy) {
        this.map = new HashMap<K, V>(mapToCopy);
      }
      this.mapToCopy = mapToCopy;
    }

    public void clear() {
      this.map.clear();
    }

    public boolean containsKey(final Object key) {
      return this.map.containsKey(key);
    }

    public boolean containsValue(final Object value) {
      return this.map.containsValue(value);
    }

    public Set<java.util.Map.Entry<K, V>> entrySet() {
      return Collections.unmodifiableSet(this.map.entrySet());
    }

    public V get(final Object key) {
      return this.map.get(key);
    }

    public boolean isEmpty() {
      return this.map.isEmpty();
    }

    public Set<K> keySet() {
      return Collections.unmodifiableSet(this.map.keySet());
    }

    public V put(final K key, final V value) {
      this.addedEntries.add(key);
      return this.map.put(key, value);
    }

    public void putAll(final Map< ? extends K, ? extends V> t) {
      for (final Entry< ? extends K, ? extends V> e : t.entrySet()) {
        this.put(e.getKey(), e.getValue());
      }
    }

    @SuppressWarnings("unchecked")
    public V remove(final Object key) {
      if (this.map.containsKey(key)) {
        this.removedEntries.add((K) key);
      }
      return this.map.remove(key);
    }

    public int size() {
      return this.map.size();
    }

    public Collection<V> values() {
      return Collections.unmodifiableCollection(this.map.values());
    }

    void synchronize() {
      synchronized (this.mapToCopy) {
        for (final K key : this.addedEntries) {
          this.mapToCopy.put(key, this.get(key));
        }
        for (final K key : this.removedEntries) {
          this.mapToCopy.remove(key);
        }
      }
    }

    @Override
    public String toString() {
      return "TXMap[added:" + this.addedEntries + ", removed " + this.removedEntries + "]";
    }
  }

  private static Map<ReplierKey, MessageCarrier> messageCarriers =
    new HashMap<ReplierKey, MessageCarrier>();

  private static Map<String, MessageCarrier> pendingMessageCarriers =
    new HashMap<String, MessageCarrier>();

  private static Map<Transaction, TransactionnalMap<ReplierKey, MessageCarrier>> msgCarriersCopies =
    new HashMap<Transaction, TransactionnalMap<ReplierKey,MessageCarrier>>();
  private static Map<Transaction, TransactionnalMap<String, MessageCarrier>> pendingMsgCarriersCopies =
    new HashMap<Transaction, TransactionnalMap<String,MessageCarrier>>();

  private MessageCarrierRepository() {
  }

  private static synchronized Map<String, MessageCarrier> getPendingMessageCarriers() {
    final StandardTransaction transaction = Environment.getCurrent().get(StandardTransaction.class);
    if (transaction != null) {
      TransactionnalMap<String, MessageCarrier> pendingMsgCarriers =
        MessageCarrierRepository.pendingMsgCarriersCopies.get(transaction);
      if (pendingMsgCarriers == null) {
        pendingMsgCarriers =
          new TransactionnalMap<String, MessageCarrier>(transaction, MessageCarrierRepository.pendingMessageCarriers);
        MessageCarrierRepository.pendingMsgCarriersCopies.put(transaction, pendingMsgCarriers);

        final StandardResource standardResource = new StandardResource() {
          public void rollback() {
            MessageCarrierRepository.pendingMsgCarriersCopies.remove(transaction);
          }

          public void prepare() {
            final TransactionnalMap<String, MessageCarrier> tm =
              MessageCarrierRepository.pendingMsgCarriersCopies.get(transaction);
            synchronized (tm.mapToCopy) {
              for (final String key : tm.addedEntries) {
                if (tm.mapToCopy.containsKey(key)) {
                  throw new OrchestraRuntimeException("Concurrent update exception: " + key + " already added to map.");
                }
              }
              for (final String key : tm.removedEntries) {
                if (!tm.mapToCopy.containsKey(key) && !tm.addedEntries.contains(key)) {
                  throw new OrchestraRuntimeException("Concurrent update exception: " + key + " already removed from map.");
                }
              }
            }
          }

          public void commit() {
            final TransactionnalMap<String, MessageCarrier> tm =
              MessageCarrierRepository.pendingMsgCarriersCopies.remove(transaction);
            tm.synchronize();
          }
        };
        transaction.enlistResource(standardResource);
      }
      return pendingMsgCarriers;
    } else {
      return MessageCarrierRepository.pendingMessageCarriers;
    }
  }

  static synchronized Map<ReplierKey, MessageCarrier> getMessageCarriers() {
    final StandardTransaction transaction = Environment.getCurrent().get(StandardTransaction.class);
    if (transaction != null) {
      TransactionnalMap<ReplierKey, MessageCarrier> msgCarriers = MessageCarrierRepository.msgCarriersCopies.get(transaction);
      if (msgCarriers == null) {
        msgCarriers = new TransactionnalMap<ReplierKey, MessageCarrier>(transaction, MessageCarrierRepository.messageCarriers);
        MessageCarrierRepository.msgCarriersCopies.put(transaction, msgCarriers);

        final StandardResource standardResource = new StandardResource() {
          public void rollback() {
            MessageCarrierRepository.msgCarriersCopies.remove(transaction);
          }

          public void prepare() {
            final TransactionnalMap<ReplierKey, MessageCarrier> tm =
              MessageCarrierRepository.msgCarriersCopies.get(transaction);
            synchronized (tm.mapToCopy) {
              for (final ReplierKey key : tm.addedEntries) {
                if (tm.mapToCopy.containsKey(key)) {
                  throw new OrchestraRuntimeException("Concurrent update exception: " + key + " already added to map.");
                }
              }
              for (final ReplierKey key : tm.removedEntries) {
                if (!tm.mapToCopy.containsKey(key) && !tm.addedEntries.contains(key)) {
                  throw new OrchestraRuntimeException("Concurrent update exception: " + key + " already removed from map.");
                }
              }
            }
          }

          public void commit() {
            final TransactionnalMap<ReplierKey, MessageCarrier> tm =
              MessageCarrierRepository.msgCarriersCopies.remove(transaction);
            tm.synchronize();
          }
        };
        transaction.enlistResource(standardResource);

      }
      return msgCarriers;
    } else {
      return MessageCarrierRepository.messageCarriers;
    }
  }

  public static synchronized void addMessageCarrier(final ReplierKey replierKey,
      final MessageCarrier messageCarrier) {
    MessageCarrierRepository.getMessageCarriers().put(replierKey, messageCarrier);
  }

  public static synchronized void associateMessage(final ReplierKey replierKey,
      final MessageVariable message) {
    final MessageCarrier messageCarrier = MessageCarrierRepository.getMessageCarriers().get(replierKey);
    if (messageCarrier == null) {
      throw new OrchestraRuntimeException("No MessageCarrier corresponding to key = " + replierKey + ".");
    }
    MessageCarrierRepository.removeMessageCarrier(replierKey);
    messageCarrier.setMessage(message);
  }

  /**
   * @param replierKey
   * @param faultQName
   * @param message
   */
  public static void associateFaultMessage(final ReplierKey replierKey, final QName faultQName, final MessageVariable message) {
    final MessageCarrier messageCarrier = MessageCarrierRepository.getMessageCarriers().get(replierKey);
    if (messageCarrier == null) {
      throw new OrchestraRuntimeException("No MessageCarrier corresponding to key = " + replierKey + ".");
    }
    MessageCarrierRepository.removeMessageCarrier(replierKey);
    messageCarrier.setFaultMessage(faultQName, message);

  }

  /**
   * Method called when there is no reply because process instance is ended.
   * @param replierKey
   */
  public static void throwMissingReply(final ReplierKey replierKey) {
    final MessageCarrier messageCarrier = MessageCarrierRepository.removeMessageCarrier(replierKey);
    if (messageCarrier == null) {
      Misc.fastDynamicLog(MessageCarrierRepository.LOG, Level.WARNING, "No MessageCarrier corresponding to key = " + replierKey
          + ", the process instance has been exited. This warning is normal after a server restart.");
    } else {
      messageCarrier.throwMissingReply();
    }
  }

  public static synchronized MessageCarrier removeMessageCarrier(final ReplierKey replierKey) {
    return MessageCarrierRepository.getMessageCarriers().remove(replierKey);
  }

  public static synchronized void addPendingMessageCarrier(
      final MessageCarrier messageCarrier) {
    MessageCarrierRepository.getPendingMessageCarriers().put(messageCarrier.getUuid(), messageCarrier);
  }

  public static synchronized MessageCarrier removePendingMessageCarrier(
      final String messageCarrierId) {
    return MessageCarrierRepository.getPendingMessageCarriers().remove(messageCarrierId);
  }

  /**
   * Method called when there is no reply because the message was not associated to any process instance.
   * @param replierKey
   */
  public static void throwMissingReply(final String messageCarrierId) {
    final MessageCarrier messageCarrier = MessageCarrierRepository.removePendingMessageCarrier(messageCarrierId);
    if (messageCarrier == null) {
      Misc.fastDynamicLog(MessageCarrierRepository.LOG, Level.WARNING,
          "No MessageCarrier corresponding to key = " + messageCarrierId
          + ". This warning is normal after a server restart.");
    } else {
      messageCarrier.throwMissingReply();
    }
  }
}
