/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.services.impl;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.wsdl.Definition;
import javax.wsdl.Port;
import javax.wsdl.Service;
import javax.xml.namespace.QName;

import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.definition.Node;
import org.ow2.orchestra.definition.activity.AbstractActivity;
import org.ow2.orchestra.definition.element.InboundMessageElement;
import org.ow2.orchestra.env.Environment;
import org.ow2.orchestra.facade.exception.DeploymentException;
import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.facade.uuid.ProcessDefinitionUUID;
import org.ow2.orchestra.services.OperationKey;
import org.ow2.orchestra.services.ReceivingElement;
import org.ow2.orchestra.services.itf.Publisher;
import org.ow2.orchestra.util.Misc;
import org.ow2.orchestra.wsdl.WsdlComparator;
import org.ow2.orchestra.wsdl.WsdlsInfos;

public class PublisherImpl implements Publisher {

  private static final Logger LOG = Logger.getLogger(PublisherImpl.class.getName());

  private final Map<ProcessDefinitionUUID, BpelProcess> bpelProcesses =
    new HashMap<ProcessDefinitionUUID, BpelProcess>();

  private final Map<String, ProcessEndpoint> processEndpoints =
    new HashMap<String, ProcessEndpoint>();

  public void publishServices(final BpelProcess bpelProcess, final Environment environment) {
    Misc.fastDynamicLog(PublisherImpl.LOG, Level.INFO,
        "Publishing services for Bpel process %s (%s).", bpelProcess.getUUID(), bpelProcess.getQName());
    this.addBpelProcess(bpelProcess);
  }

  public void unpublishServices(final BpelProcess bpelProcess, final Environment environment) {
    Misc.fastDynamicLog(PublisherImpl.LOG, Level.INFO,
        "Unpublishing services for Bpel process %s (%s).", bpelProcess.getUUID(), bpelProcess.getQName());
    this.removeBpelProcess(bpelProcess);
  }

  public Set<OperationKey> getServiceOperationKeys(final BpelProcess bpelProcess, final Environment environment) {
    return this.getServiceOperationKeys(bpelProcess, bpelProcess.getInitial(), environment);
  }

  protected Set<OperationKey> getServiceOperationKeys(final BpelProcess bpelProcess, final Node node,
      final Environment environment) {

    final Set<OperationKey> operationKeys = new HashSet<OperationKey>();
    // get list of operation keys for the current node
    final AbstractActivity behaviour = node.getBehaviour();
    if (behaviour.isInboundMessageElement()) {
      final InboundMessageElement inboundMessageElement = (InboundMessageElement) behaviour;
      for (final ReceivingElement receivingElement : inboundMessageElement.getReceivingElements()) {
        final OperationKey operationKey = receivingElement.getOperationKey();
        operationKeys.add(operationKey);
      }
    }
    // Add list of operation keys defined in children nodes
    final List<Node> children = node.getNodes();
    if (children != null) {
      for (final Node child : children) {
        operationKeys.addAll(this.getServiceOperationKeys(bpelProcess, child, environment));
      }
    }
    return operationKeys;
  }

  public Set<Definition> getWsdlDefinitions(final ProcessDefinitionUUID processDefinitionUUID) {
    if (this.bpelProcesses.containsKey(processDefinitionUUID)) {
      return this.bpelProcesses.get(processDefinitionUUID).getWsdlInfos().getWsdlDefinitions();
    }
    return null;
  }

  public BpelProcess getBpelProcess(final ProcessDefinitionUUID processDefinitionUUID) {
    return this.bpelProcesses.get(processDefinitionUUID);
  }


  public void addBpelProcess(final BpelProcess bpelProcess) {
    this.bpelProcesses.put(bpelProcess.getUUID(), bpelProcess);
  }

  public void removeBpelProcess(final BpelProcess bpelProcess) {
    this.bpelProcesses.remove(bpelProcess.getUUID());
  }

  /**
   * Register endpoint for the version of the process
   * @return true if no other version of the process use the endpoint (and the web service has to be deployed)
   */
  public boolean addProcessEndpoint(
      final String endpoint,
      final QName processQName,
      final Port port,
      final ProcessDefinitionUUID processDefinitionUUID,
      final Map<String, Boolean> locks,
      final Map<String, Boolean> isOneWay) {
    final boolean isNewEndpoint = !this.processEndpoints.containsKey(endpoint);
    final ProcessEndpoint processEndpoint;
    if (isNewEndpoint) {
      processEndpoint = new ProcessEndpoint(processQName, port);
      this.processEndpoints.put(endpoint, processEndpoint);
    } else {
      processEndpoint = this.processEndpoints.get(endpoint);
      WsdlComparator.checkSamePort(processEndpoint.getPort(), port);
      if (!processEndpoint.getProcessQName().equals(processQName)) {
        throw new DeploymentException(
            "endpoint " + endpoint + " is already used by process " + processEndpoint.getProcessQName());
      }
    }
    processEndpoint.addProcessVersion(processDefinitionUUID, isOneWay, locks);
    return isNewEndpoint;
  }

  /**
   * Unregister endpoint for the version of the process.
   * Deletes messages given to this endpoint but not consumed by any process instance.
   * @param endpoint
   * @param processDefinitionUUID
   * @return true if no more version of the process use the endpoint (and the web service can be removed)
   */
  public boolean removeProcessEndpoint(final String endpoint, final ProcessDefinitionUUID processDefinitionUUID) {
    final ProcessEndpoint processEndpoint = this.processEndpoints.get(endpoint);
    if (processEndpoint != null) {
      processEndpoint.removeProcessVersion(processDefinitionUUID);
      if (processEndpoint.isUndeployed()) {
        this.processEndpoints.remove(endpoint);
        return true;
      }
    }
    return false;
  }

  public ProcessEndpoint getProcessEndpoint(final String endpoint) {
    return this.processEndpoints.get(endpoint);
  }

  protected List<Service> getServices(final BpelProcess bpelProcess) {
    // get operation keys for receiving activities
    final Collection<OperationKey> opKeys = this.getServiceOperationKeys(bpelProcess, Environment.getCurrent());

    final WsdlsInfos wsdlInfos = bpelProcess.getWsdlInfos();
    List<Service> services = null;

    // to store portTypes of receving activities that does not have a service associated
    final List<QName> missingServices = new ArrayList<QName>();

    // try to find a WSDL service to deploy for each receiving activity
    for (final OperationKey operationKey : opKeys) {
      final QName portTypeQName = operationKey.getPortTypeQName();
      final List<Service> portTypeServices = wsdlInfos.getServicesOfPortType(portTypeQName);
      if (portTypeServices != null) {
        if (services == null) {
          services = new ArrayList<Service>();
        }
        for (final Service service : portTypeServices) {
          if (!services.contains(service)) {
            services.add(service);
          }
        }
      } else {
        // no service found for port type
        missingServices.add(portTypeQName);
      }
    }

    // throw exception if there are missing services.
    if (!missingServices.isEmpty()) {
      final StringBuffer buf = new StringBuffer();
      buf.append("Some wsdl services not found for receiving activities. Please specify service for the following PortTypes: ");
      boolean first = true;
      for (final QName ptQName : missingServices) {
        buf.append(first ? "" : ", ").append(ptQName.toString());
        first = false;
      }
      buf.append(".");
      throw new OrchestraRuntimeException(buf.toString());
    }
    return services;
  }
}
