/*
 * JBoss, Home of Professional Open Source
 * Copyright 2005, JBoss Inc., and individual contributors as indicated
 * by the @authors tag. See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.ow2.orchestra.services.jobexecutor;

import java.util.logging.Level;
import java.util.logging.Logger;

import org.ow2.orchestra.env.Environment;
import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.persistence.Transaction;
import org.ow2.orchestra.persistence.db.MessageSession;
import org.ow2.orchestra.services.job.MessageImpl;
import org.ow2.orchestra.util.Misc;
import org.ow2.orchestra.util.ReflectUtil;

/**
 * @author Tom Baeyens
 */
public class JobExecutorMessageSession implements MessageSession {

  private static final Logger LOG = Logger.getLogger(JobExecutorMessageSession.class.getName());

  /* injected */
  private JobDbSession jobDbSession;

  /* injected */
  private Transaction transaction;

  /* injected */
  private int defaultRetries = 3;

  /* injected */
  private boolean useFairScheduling = true;

  public void send(final MessageImpl< ? > message) {
    Misc.fastDynamicLog(JobExecutorMessageSession.LOG, Level.FINE,
        "sending message %s", ReflectUtil.getUnqualifiedClassName(message.getClass()));

    if (message.isDefaultRetries()) {
      message.setRetries(this.defaultRetries);
    }
    if (!this.useFairScheduling
        && message.getExecution() != null
        && message.getDueDate().getTime() <= System.currentTimeMillis()) {
      message.setDueDate(message.getExecution().getProcessInstance().getParentInstanceStartTime());
    }
    this.jobDbSession.save(message);

    // notify the job executor after the transaction is completed
    Misc.fastDynamicLog(JobExecutorMessageSession.LOG, Level.FINEST,
        "registering job executor notifier with %s", this.transaction);
    final JobExecutor jobExecutor = Environment.getCurrent().get(JobExecutor.class);
    if (jobExecutor == null) {
      throw new OrchestraRuntimeException("no job executor configured to execute "
          + message);
    }
    this.transaction
        .registerSynchronization(new JobAddedNotification(jobExecutor, message));
  }


  /**
   * @param defaultRetries the defaultRetries to set
   */
  public void setDefaultRetries(final int defaultRetries) {
    this.defaultRetries = defaultRetries;
  }


  /**
   * @param useFairScheduling the useFairScheduling to set
   */
  public void setUseFairScheduling(final boolean useFairScheduling) {
    this.useFairScheduling = useFairScheduling;
  }
}
