/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.util;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.xml.namespace.QName;

import org.ow2.orchestra.definition.activity.AbstractActivity;
import org.ow2.orchestra.definition.element.Copy;
import org.ow2.orchestra.definition.element.Correlation;
import org.ow2.orchestra.definition.element.CorrelationSet;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.runtime.CorrelationSetRuntime;
import org.ow2.orchestra.runtime.ScopeRuntime;
import org.ow2.orchestra.runtime.VariableRuntime;
import org.ow2.orchestra.services.MessageCarrier;
import org.ow2.orchestra.services.Receiver;
import org.ow2.orchestra.services.ReceivingElement;
import org.ow2.orchestra.var.MessageVariable;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;


/**
 * @author Guillaume Porcher
 *
 */
public final class ReceivingElementUtil {

  private ReceivingElementUtil() { }

  private static final Logger LOG = Logger.getLogger(ReceivingElementUtil.class.getName());

  /**
   * This message is to be called from the receive implementation (above) in
   * case the message arrived in the engine before the execution arrived in the
   * receive. in case the message arrives later, the receive will be invoked by
   * the correlator.
   */
  public static void receive(
      final ReceivingElement receivingElement,
      final MessageVariable message,
      final MessageCarrier messageCarrier,
      final BpelExecution execution) {
    Misc.fastDynamicLog(ReceivingElementUtil.LOG, Level.FINE,"Receiving message : %s", message);

    // initiate correlationSets
    ReceivingElementUtil.initiateCs(receivingElement, execution, message);
    if (receivingElement.getVariable() != null) {
      // we store the variable
      final VariableRuntime variableRuntime = execution.getVariableByName(receivingElement.getVariable());
      variableRuntime.setValue(message);
      // record variable updated
      final AbstractActivity activity = (AbstractActivity) execution.getNode().getBehaviour();
      variableRuntime.updated(activity.createRuntimeData(execution));
    } else if (receivingElement.getInCopies() != null && !receivingElement.getInCopies().isEmpty()) {
      final ScopeRuntime scopeRuntime = execution.getCurrentScopeRuntime();
      final VariableRuntime anonymousInboundMessageVariableRuntime = scopeRuntime
          .createVariable(execution, receivingElement.getAnonymousInboundMessageVariable());

      anonymousInboundMessageVariableRuntime.setValue(message);
      for (final Copy copy : receivingElement.getInCopies()) {
        copy.execute(execution);
      }
      scopeRuntime.removeVariable(execution,
          receivingElement.getAnonymousInboundMessageVariable());
    }

    // Add replier if needed
    Receiver.addReplier(execution, receivingElement, messageCarrier);
  }

  private static void initiateCs(
      final ReceivingElement receivingElement,
      final BpelExecution bpelExecution,
      final MessageVariable incomingMessage) {
    if (receivingElement.getCorrelations() != null) {
      // we calcul the values for each correlationSet
      for (final Correlation correlation : receivingElement.getCorrelations()) {
        final CorrelationSetRuntime correlationSetRuntime =
          bpelExecution.getCorrelationSetByName(correlation.getCorrelationSet().getName());
        final QName messageType = receivingElement.getInputMessageType();
        correlationSetRuntime.checkCSValues(bpelExecution, incomingMessage, messageType, correlation);
      }
    }
  }

  public static void initiateCs(final ReceivingElement receivingElement,
      final BpelExecution bpelExecution,
      final CorrelationSet cs,
      final MessageVariable incomingMessage) {
    if (receivingElement.getCorrelations() != null) {
      // we calcul the values for each correlationSet
      for (final Correlation correlation : receivingElement.getCorrelations()) {
        if (correlation.getCorrelationSet().equals(cs)) {
          final CorrelationSetRuntime correlationSetRuntime =
            bpelExecution.getCorrelationSetByName(correlation.getCorrelationSet().getName());
          final QName messageType = receivingElement.getInputMessageType();
          correlationSetRuntime.checkCSValues(bpelExecution, incomingMessage, messageType, correlation);
        }
      }
    }
  }

  public static int getCorrelationPropertiesKey(final Map<QName, Document> csProperties) {
    // Sort correlation properties by qname.
    final List<QName> sortedCsNames = new ArrayList<QName>(csProperties.keySet());
    Collections.sort(sortedCsNames, QNameComparator.INSTANCE);
    // For each correlation property, generate a string form property value
    final StringBuilder sb = new StringBuilder();
    for (final QName csQnName : sortedCsNames) {
      sb.append(csQnName.toString());
      ReceivingElementUtil.getNormalizedStringFromXML(csProperties.get(csQnName), sb);
    }
    final int hashCode = sb.toString().hashCode();
    Misc.fastDynamicLog(ReceivingElementUtil.LOG, Level.FINE, "Generating hash code for correlation properties: %s",
        Integer.valueOf(hashCode));
    return hashCode;
  }

  private static void getNormalizedStringFromXML(final Node node, final StringBuilder stringBuilder) {
    if (node instanceof Element) {
      stringBuilder.append('<').append(node.getLocalName()).append('>');
      final NodeList nl = node.getChildNodes();
      for (int i = 0; i < nl.getLength(); i++) {
        ReceivingElementUtil.getNormalizedStringFromXML(nl.item(i), stringBuilder);
      }
      stringBuilder.append("</").append(node.getLocalName()).append('>');
    } else if (node instanceof Text) {
      stringBuilder.append(node.getTextContent().trim());
    } else if (node instanceof Document) {
      ReceivingElementUtil.getNormalizedStringFromXML(((Document) node).getDocumentElement(), stringBuilder);
    }
  }
}
