/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.var;

import java.util.Map;
import java.util.Map.Entry;

import javax.wsdl.Part;
import javax.xml.namespace.QName;

import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.facade.runtime.impl.MessageImpl;
import org.ow2.orchestra.util.BpelXmlUtil;
import org.ow2.orchestra.util.XmlUtil;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

/**
 * Models a WSDL message type variable.
 *
 * @author Goulven Le Jeune, Charles Souillard, Candy Chlad, Stanislas Giraudet
 *         De Boudemange
 */
public class MessageVariable extends MessageImpl {

  private static final long serialVersionUID = -3214634040146001613L;

  /**
   * Empty message constructor. For hibernate.
   */
  protected MessageVariable() { }

  /**
   * Create a multi-part message.
   *
   * @param parts
   *            HashMap containing the message parts
   */
  public MessageVariable(final Document parts) {
    this.parts = parts;
  }

  /**
   * Constructor based on the wsdl Message.
   *
   * @param wsdlMessage
   *            wsdlMessage that specify this message
   */
  @SuppressWarnings("unchecked")
  public MessageVariable(final javax.wsdl.Message wsdlMessage) {
    this.parts = XmlUtil.getNewDocument();
    this.parts.appendChild(this.parts.createElement("message"));

    final Map<String, Part> wsdlMessageParts = wsdlMessage.getParts();
    for (final Entry<String, Part> entry : wsdlMessageParts.entrySet()) {
      final Part part = entry.getValue();
      final String partName = part.getName();
      // this.parts.put(partName, null);
      final QName elementQName = part.getElementName();
      final QName typeQName = part.getTypeName();
      final Element partElement = (Element) this.parts.getDocumentElement().appendChild(this.parts.createElement(partName));
      if (typeQName != null) {
        partElement.appendChild(BpelXmlUtil.createDefaultElement(this.parts));
      } else {
        final Element partContent;
        if (elementQName.getPrefix() != null && !elementQName.getPrefix().equals("")) {
          partContent = this.parts.createElementNS(
            elementQName.getNamespaceURI(),
            elementQName.getPrefix() + ":" + elementQName.getLocalPart());
        } else {
          partContent = this.parts.createElementNS(
              elementQName.getNamespaceURI(),
              elementQName.getLocalPart());
        }
        partElement.appendChild(partContent);
        BpelXmlUtil.ensureOwnNamespaceDeclared(partContent);
      }
    }
  }

  /**
   * Return true if the part having the given name is initialized.
   *
   * @param partName
   *            partName
   * @return true if the part having the given name is initialized.
   */
  public boolean isPartInitialized(final String partName) {
    return this.getPartValue(partName) != null;
  }

  /**
   * Return a string representation of the message.
   *
   * @return a String representation of the message
   */
  @Override
  public String toString() {
    return XmlUtil.toString(this.parts);
  }


  /**
   * Make a deep copy of the Xml object.
   *
   * @return The duplicate of Xml object
   */
  public MessageVariable duplicate() {
    if (this.parts == null) {
      return null;
    }
    final Document duplicateParts = XmlUtil.getNewDocument();
    final Node n = duplicateParts.importNode(this.parts.getDocumentElement(), true);
    duplicateParts.appendChild(n);
    return new MessageVariable(duplicateParts);
  }

  public Document getPartsAsDocument() {
    return this.parts;
  }

  /**
   * Compare the Message to another one.
   *
   * @param anotherMessage :
   *            the other Message to compare to
   * @return true if all parts are equals, else false
   */
  @Override
  public boolean equals(final Object anotherMessage) {
    if (anotherMessage == null) {
      return false;
    }
    if (!(anotherMessage instanceof MessageVariable)) {
      return false;
    }
    final MessageVariable message = (MessageVariable) anotherMessage;

    return XmlUtil.areNodeEquals(this.parts, message.parts);
  }

  /**
   * Returns a hash code value for the Message.
   *
   * @return a hash code value for this Message.
   */
  @Override
  public int hashCode() {
    return this.parts.hashCode();
  }

  public void setPart(final String partName, final Object value, final boolean keepSrcElementName) {
    if (!this.hasPart(partName)) {
      throw new OrchestraRuntimeException("Invalid part name: " + partName);
    }
    final Element currentValue = this.getPartValue(partName);
    BpelXmlUtil.setObjectValue(currentValue, value, keepSrcElementName);
  }
}
